/*
 *   Copyright (C) 2005 Sergey Koposov
 * 
 * Author: Sergey Koposov, Sternberg Astronomical Institute
 * Email: math@sai.msu.ru 
 * http://lnfm1.sai.msu.ru/~math

  This file is part of GMATCH.

  GMATCH is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  GMATCH is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with GMATCH; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */


#include <string.h>
#include <math.h>
#include "fitsio.h"
#include "wcs.h"
#include "config.h"

#define __FERR fits_report_error(stderr, *status); if (*status) return(*status)



void init_arrays1(int num, double **radec, double **phi, double **theta,
					double **imgcrd, int **stat);



int fix_wcs(char * filename, double *cd, double *crpix, double *crval)
{
	fitsfile *fptr;
	int status[1] = {0};


	fits_open_image(&fptr, filename, READWRITE, status);
	__FERR;

	fits_update_key_lng(fptr, "WCSAXES", 2, "", status);
	__FERR;

	fits_update_key_dbl(fptr, "CRPIX1", crpix[0], 10, "", status);
	__FERR;

	fits_update_key_dbl(fptr, "CRPIX2", crpix[1], 10, "", status);
	__FERR;

	fits_update_key_dbl(fptr, "CRVAL1", crval[0], 10, "", status);
	__FERR;

	fits_update_key_dbl(fptr, "CRVAL2", crval[1], 10, "", status);
	__FERR;


	fits_update_key_str(fptr, "RADECSYS", "ICRS", "", status);
	__FERR;

	fits_update_key_dbl(fptr, "EPOCH", 2000, 0,"", status);
	__FERR;
	
	/* Very temporary and due to bug in DS9 (DS9 uses the EPOCH keyword even 
	 * when RADECSYS='ICRS' and some frames (like Zeiss1000 images) have the 
	 * epoch of the observation in the EPOCH keyword
	 */


	fits_update_key_str(fptr, "CTYPE1", "RA---TAN", "", status);
	__FERR;

	fits_update_key_str(fptr, "CTYPE2", "DEC--TAN", "", status);
	__FERR;

	fits_update_key_str(fptr, "CUNIT1", "deg    ", "", status);
	__FERR;

	fits_update_key_str(fptr, "CUNIT2", "deg    ", "", status);
	__FERR;

	fits_update_key_dbl(fptr, "CD1_1", cd[0], 10, "", status);
	__FERR;

	fits_update_key_dbl(fptr, "CD1_2", cd[1], 10, "", status);
	__FERR;

	fits_update_key_dbl(fptr, "CD2_1", cd[2], 10, "", status);
	__FERR;

	fits_update_key_dbl(fptr, "CD2_2", cd[3], 10, "", status);
	__FERR;

	
	/* Inserted these lines only for MIDAS since it doesn't 
	 * like CDELTi be present together with CDi_j
	 */
	fits_delete_key(fptr, "CDELT1", status);
	*status=0;
		
	fits_delete_key(fptr, "CDELT2", status);
	*status=0;
			  
//  fits_update_key_dbl(fptr, "CDELT1", 1, 10, "", status);
//  __FERR;

//  fits_update_key_dbl(fptr, "CDELT2", 1, 10, "", status);
//  __FERR;

	fits_close_file(fptr, status);
	return *status;

}


/*
 * @return	Returns 0 if OK, and FITSIO error status in contrary case
 * @param filename 	The name of the FITS file 
 *
 * @param size	The preallocated array with two elements which will contain 
 *           	the pixel size of the image
 *
 */
int gmatch_get_image_size(char *filename, long *size)
{
	int status[1]={0};
	fitsfile *fptr;
	
	fits_open_image(&fptr, filename, READONLY, status);
	__FERR;
	fits_get_img_size(fptr, 2, size, status);
	__FERR;
	fits_close_file(fptr, status);
	return *status;
}

int fits_wcs_init(char *filename, struct wcsprm *our_wcs, long *size)
{

	int status[1] = {0};
	double crpix1, crpix2, cd11, cd12, cd21, cd22, pc11, pc12, pc21, pc22,
			   crval1, crval2, cdelt1 = 1 , cdelt2 = 1, lonpole = 999, latpole = 999,
			   equinox = 2000, crota1, crota2;
	char cd_found = 0, pc_found = 0, cdelt_found = 0, crpix_found = 0,
			 crval_found = 0, crota_found=0, cunit_found=0, ctype_found=0 ;
	char cunit1[80]="deg", cunit2[80]="deg";
	char ctype1[80]="RA---TAN", ctype2[80]="DEC--TAN";
	char radesys[80] = "ICRS";
	fitsfile *fptr;

	fits_open_image(&fptr, filename, READONLY, status);
	__FERR;
	
	memset(our_wcs,0,sizeof(our_wcs));
	our_wcs->flag = -1;   
	wcsini(1, 2, our_wcs);
	our_wcs->flag = 0; 
	our_wcs->naxis = 2;

	*status=0;
	fits_read_key_dbl(fptr,"CRPIX1",&crpix1,NULL,status);
	fits_read_key_dbl(fptr,"CRPIX2",&crpix2,NULL,status);

	if (*status==0) crpix_found=1;

	if (crpix_found)
	{
		our_wcs->crpix[0] = crpix1;
		our_wcs->crpix[1] = crpix2;
	}

	*status=0;
	fits_read_key_dbl(fptr,"CRVAL1",&crval1,NULL,status);
	fits_read_key_dbl(fptr,"CRVAL2",&crval2,NULL,status);

	if (*status==0) crval_found=1;

	if (crval_found)
	{
		our_wcs->crval[0] = crval1;
		our_wcs->crval[1] = crval2;
	}


	*status=0;
	fits_read_key_dbl(fptr,"CD1_1",&cd11,NULL,status);
	fits_read_key_dbl(fptr,"CD2_1",&cd21,NULL,status);
	fits_read_key_dbl(fptr,"CD1_2",&cd12,NULL,status);
	fits_read_key_dbl(fptr,"CD2_2",&cd22,NULL,status);

	if (*status==0) cd_found=1;

	if (cd_found) 
	{
		our_wcs->cd[0] = cd11;           
		our_wcs->cd[1] = cd12;    
		our_wcs->cd[2] = cd21;
		our_wcs->cd[3] = cd22;
		our_wcs->altlin|=2;
	}
	

	
	*status=0;
	fits_read_key_dbl(fptr,"PC1_1",&pc11,NULL,status);
	fits_read_key_dbl(fptr,"PC2_1",&pc21,NULL,status);
	fits_read_key_dbl(fptr,"PC1_2",&pc12,NULL,status);
	fits_read_key_dbl(fptr,"PC2_2",&pc22,NULL,status);

	if (*status==0) pc_found=1;
	
	if (pc_found) 
	{
		our_wcs->pc[0] = pc11;           
		our_wcs->pc[1] = pc12;    
		our_wcs->pc[2] = pc21;
		our_wcs->pc[3] = pc22;
		our_wcs->altlin|=1;
	}

	*status=0;
	fits_read_key_dbl(fptr,"CROTA1",&crota1,NULL,status);
	fits_read_key_dbl(fptr,"CROTA2",&crota2,NULL,status);

	if (*status==0) crota_found=1;

	if (crota_found) 
	{
		our_wcs->crota[0] = crota1;           
		our_wcs->crota[1] = crota2;    
		our_wcs->altlin |= 4;    
	}

	*status=0;  
	fits_read_key_dbl(fptr,"CDELT1",&cdelt1,NULL,status);
	fits_read_key_dbl(fptr,"CDELT2",&cdelt2,NULL,status);

	if (*status==0) cdelt_found=1;

	our_wcs->cdelt[0] = cdelt1;           
	our_wcs->cdelt[1] = cdelt2;    
	

	*status=0;
	fits_read_key_str(fptr,"CUNIT1",cunit1,NULL,status);
	fits_read_key_str(fptr,"CUNIT2",cunit2,NULL,status);

	if (*status==0) cunit_found=1;

	strcpy((our_wcs->cunit)[0],cunit1);
	strcpy((our_wcs->cunit)[1],cunit2);
	
	*status=0;
	fits_read_key_str(fptr,"CTYPE1",ctype1,NULL,status);
	fits_read_key_str(fptr,"CTYPE2",ctype2,NULL,status);
	if (*status==0) ctype_found=1;

	strcpy((our_wcs->ctype)[0],ctype1);
	strcpy((our_wcs->ctype)[1],ctype2);
	
	fits_read_key_str(fptr,"RADECSYS",radesys,NULL,status);
	strcpy(our_wcs->radesys,radesys);

	*status=0;
	fits_read_key_dbl(fptr,"LONPOLE",&lonpole,NULL,status);
	*status=0;
	fits_read_key_dbl(fptr,"LATPOLE",&latpole,NULL,status);
	*status=0;
	fits_read_key_dbl(fptr,"EQUINOX",&equinox,NULL,status);

	our_wcs->lonpole = lonpole;    
	our_wcs->latpole = latpole;  
	our_wcs->equinox = equinox;  
	//        our_wcs->equinox = 2000.;
	wcsset(our_wcs);

//  fits_read_key_lng(fptr,"NPIX1",size, NULL,status);

	*status=0;
	fits_get_img_size(fptr, 2, size, status);
	
	fits_close_file(fptr, status);
	return 0;

}


double sph_dist(double ra1, double dec1, double ra2, double dec2)
{
	double c = M_PI/180;
	double sd=sin(c * (dec1 - dec2) / 2);
	double cd=cos(c * (dec1 + dec2) / 2);
	double sr=sin(c * (ra1 - ra2));
	sr *= sr;
	cd *= cd;
	sd *= sd;
	c = 180 / M_PI;
	return c * 2 * asin( sqrt ( sr * (cd - sd ) + sd  ) ) ;
}


void get_fits_cen(char *filename, double *ra_cen, double  *dec_cen,
					double *im_rad)
{
	long size[2];
	struct wcsprm our_wcs;
	double xy[10], *phi, *theta, *imgcrd, *radec, dist, dist1;
	int *stat, npoints, i;

	fits_wcs_init(filename, &our_wcs, size);
	
	npoints = 5;
	xy[0] = (size[0]+1.)/2.;
	xy[1] = (size[1]+1.)/2.;
	xy[2] = 1;
	xy[3] = 1;
	xy[4] = 1;
	xy[5] = size[1];
	xy[6] = size[0];
	xy[7] = size[1];
	xy[8] = size[0];
	xy[9] = 1;
		
	init_arrays1(npoints, &radec, &phi, &theta, &imgcrd, &stat);
	wcsp2s(&our_wcs, npoints, 2, xy, imgcrd, phi, theta, radec, stat);
	free(phi);
	free(theta);
	free(stat);
	free(imgcrd);
	
	*ra_cen = radec[0];
	
	*dec_cen = radec[1];
	
	dist = 0;
	for(i = 1; i <= 4; i++)
	{
		dist1 = sph_dist(radec[0], radec[1], radec[2 * i], radec[2 * i + 1]);
		dist = dist1 > dist ? dist1 : dist;
	}
	*im_rad = dist;
	free(radec);
}
