/*
    Copyright (C) 2005 Sergey Koposov
  
  Author: Sergey Koposov, Sternberg Astronomical Institute
  Email: math@sai.msu.ru 
  http://lnfm1.sai.msu.ru/~math

  This file is part of GMATCH.

  GMATCH is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  GMATCH is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with GMATCH; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */


#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include "wcs.h"
#include "common.h"
#include "common_wcs.h"
#include "config.h"

#define __FERR fits_report_error(stderr, status)
#define R_MAX_THRESH 10 

#define TR_OR(dx1, dy1, dx2, dy2) ((dx1 * dy2 - dy1 * dx2) > 0)

void matrix_multiply(gr_matrix m1, gr_matrix m2,
						gr_matrix *m)
{
	m->t[0] = m1.t[0] * m2.t[0] + m1.t[1] * m2.b[0];
	m->t[1] = m1.t[0] * m2.t[1] + m1.t[1] * m2.b[1];
	m->b[0] = m1.b[0] * m2.t[0] + m1.b[1] * m2.b[0];
	m->b[1] = m1.b[0] * m2.t[1] + m1.b[1] * m2.b[1];
}

gr_matrix matrix_invert(gr_matrix m0)
{
	gr_matrix m;
	
	gr_type det = m0.t[0] * m0.b[1] - m0.t[1] * m0.b[0];

	m.t[0] = m0.b[1] / det;
	m.b[1] = m0.t[0] / det;
	m.t[1] = -m0.t[1] / det;
	m.b[0] = -m0.b[0] / det;
	return m;  
}

gr_vector matrix_vector_multiply(gr_matrix m,
										gr_vector v0)
{
	gr_vector v;
	v.t[0] = m.t[0] * v0.t[0] + m.t[1] * v0.t[1];
	v.t[1] = m.b[0] * v0.t[0] + m.b[1] * v0.t[1];
	return v;
}

gr_vector vector_subtract(gr_vector v0, gr_vector v1)
{
	gr_vector v;
	v.t[0] = v0.t[0] - v1.t[0];
	v.t[1] = v0.t[1] - v1.t[1];
	return v;
}

void matrix_print(FILE *f, gr_matrix m)
{
	fprintf(f, "%.10f %.10f\n%.10f %.10f\n", m.t[0], m.t[1], m.b[0], m.b[1]);
}

void vector_print(FILE *f, gr_vector v)
{
	fprintf(f, "%.10f %.10f\n", v.t[0], v.t[1]);  
}


inline int gr_type_compare(const void *a, const void *b)
{
	gr_type aa = *((gr_type *)a);
	gr_type bb = *((gr_type *)b);
	if (aa < bb)
	{
		return -1;
	}
	else
	{
		if (aa == bb)
		{
			return 0;
		}
		else 
		{
			return 1;
		}
	}
}



inline int three_ints_compare(const void *a, const void *b)
{
	return -(((struct three_ints *)a)->l1 - ((struct three_ints *)b)->l1);
}

inline int three_ints_compare1(const void *a, const void *b)
{
	int t2 = (((struct three_ints *)a)->l2 - ((struct three_ints *)b)->l2);
	int t3 = (((struct three_ints *)a)->l3 - ((struct three_ints *)b)->l3);
	if (t2<0)
	{
		return -1;
	}
	else
	{
		if (t2 == 0)
		{
			return t3;
		}
		else 
		return 1;
	}
}

inline int double_and_char_compare(const void *a, const void *b)
{
	double aa = ((double_and_char *)a)->ratio ;
	double bb = ((double_and_char *)b)->ratio ;
	if (aa < bb)
	{
		return -1;
	}
	else
	{
		if (aa == bb)
		{
			return 0;
		}
		else 
		{
			return 1;
		}
	}

}



inline int ipix_compare(const void *a , const void * b)
{
	unsigned int i1 = ((ipix_ptr *)a)->ipix;
	unsigned int i2 = ((ipix_ptr *)b)->ipix;

	if (i1 > i2)
	{
		return 1;
	}
	else
	{
		if (i1 < i2)
		{
			return -1;
		}
		else 
		{
			return 0;
		}
	}
}


void get_regression(gr_type a1, gr_type ax, gr_type ay, gr_type ax1,
					gr_type axx, gr_type axx1, gr_type axy, gr_type ayx1,
					gr_type ayy, gr_type *A, gr_type *B, gr_type *C)
{
/* Thanks to Maple ;) */
gr_type      t1 = ay * ay;
gr_type      t3 = ay * axy;
gr_type      t5 = ay * ax;
gr_type      t7 = ayy * a1;
gr_type      t9 = ayy * ax;
gr_type      t11 = axy * a1;
gr_type      t18 = ax * ax;
gr_type      t20 = axy * axy;
gr_type      t23 = 1 / (-axx * t1 + 2.0 * ax * t3 + axx * t7 - ayy * t18 -
					 a1 * t20);
gr_type      t26 = axy * ax;
			*A = t23 * (-axx1 * t1 + ax1 * t3 + ayx1 * t5 + axx1 * t7 - ax1 * t9 -
				ayx1 * t11);
			*B = -t23 * (-axx1 * t5 - ax1 * t26 + ayx1 * t18 + axx1 * t11 - a1 *
				ayx1 * axx + ax1 * ay * axx);
			*C = -t23 * (-ayy * ax1 * axx + axx1 * t9 - axx1 * t3 + ay * ayx1 * axx
				-ayx1 * t26 + t20 * ax1);
}


void get_regression2(gr_type a1, gr_type ax, gr_type ay, gr_type ax1,
					gr_type ay1, gr_type axx, gr_type axx1, gr_type axy1,
					gr_type ayy, gr_type ayx1, gr_type ayy1, gr_type *A,
					gr_type *B, gr_type *C, gr_type *D)
{
/* Thanks to Maple ;) */
	gr_type t8 = ay*ay;
	gr_type t9 = ax*ax;
	gr_type t11 = 1./(a1*axx+a1*ayy-t8-t9);
	*A = t11*(-ay*ay1+a1*axx1+a1*ayy1-ax*ax1);
	*B = t11*(-ax1*ay+a1*ayx1+ax*ay1-a1*axy1);
	*C = -t11*(-axx*ax1-ayy*ax1+ay*ayx1-ay*axy1+ax*axx1+ax*ayy1);
	*D = -t11*(-ay1*axx+ay*axx1+ay*ayy1-ax*ayx1+ax*axy1-ay1*ayy);
}


void get_regression2_1(gr_type a1, gr_type ax, gr_type ay, gr_type ax1,
							       gr_type ay1, gr_type axx, gr_type axx1, gr_type axy1,
							       gr_type ayy, gr_type ayx1, gr_type ayy1, gr_type *A,
							       gr_type *B, gr_type *C, gr_type *D)
{
/* Thanks to Maple ;) */
	gr_type t6 = ax*ax;
	gr_type t7 = ay*ay;
	gr_type t11 = 1/(-t6-t7+a1*axx+a1*ayy);
	*A  = -t11*(-ay*ay1-a1*axx1+a1*ayy1+ax1*ax);
	*B = t11*(-ax*ay1+a1*ayx1-ay*ax1+a1*axy1);
	*C = t11*(axx*ax1+ayy*ax1-ayx1*ay-axy1*ay-ax*axx1+ax*ayy1);
	*D = -t11*(-ay*axx1+ax*ayx1+ax*axy1-axx*ay1-ayy*ay1+ay*ayy1);
}


/* This function compute the mode based on the histogram of the data.
 * The input array is "data" with "ndata" elements. That array should be 
 * previously sorted in ascending order
 * The histogram with 2*"nhist"+1 bins is created in the pointer "hist"
 * The pointer hist should be 0 initialized, preallocated array 
 * (having the size = 2*nhist+1)
 * posl_out and posr_out -- output parameters. This parameters delimit the
 * part of the array of around the mode (such part that between posl_out and
 * posr_out the half of the original range is contained
 */
void mode_hist(gr_type *data, int ndata, int *hist, int nhist, int *posl_out,
							int *posr_out)
{
	/* hist should be initialized array of length = 2*nhist+1 */
	gr_type min=data[0];
	gr_type max=data[ndata - 1];
	int i, pos, len, val, posl, posr;
	gr_type vall, valr;
	gr_type delta = (max - min) / (nhist);
	gr_type mult = 2 / delta;
	gr_type diff;
	for(i = 0; i < ndata; i++)
	{
		diff = data[i] - min;
		pos=(int)(diff * mult);
		hist[pos] += 1;
		if (pos > 0)
		{
			hist[pos - 1] += 1;
		}
	}
	len = 2 * nhist + 1 ;
	val = -1;
	pos=0;
	for (i = 0; i < len; i++)
	{
		if (hist[i] > val) 
		{
			pos=i;
			val=hist[i];
		}
	}
	vall = min+ pos * delta/2;
	valr = vall + (max - min)/4;
	vall = vall - (max - min)/4;
	posl = 0;
	posr = 0;
	for(i = 0; i < ndata; i++)
	{
		if (data[i] < vall)
		{
			posl++;
		}
		if (data[i] < valr)
		{
			posr++;
		}
	}
	posl++;
	*posl_out = posl;
	*posr_out = posr;  
	
}

/* This function compute the mode based on the histogram of the data.
 * The input array is "data" with "ndata" elements. That array should be 
 * previously sorted in ascending order
 * The histogram with 2*"nhist"+1 bins is created in the pointer "hist"
 * The pointer hist should be 0 initialized, preallocated array 
 * (having the size = 2*nhist+1)
 * posl_out and posr_out -- output parameters. This parameters delimit the
 * part of the array of around the mode (such part that between posl_out and
 * posr_out the half of the original range is contained
 *
 * The version with the flag
 * If the flag ==1 then we add one to the histogram bin, if flag==0 then we 
 * subtract 1. 
 */
void mode_hist1(double_and_char *data, int ndata, int *hist,
				int nhist, int *posl_out, int *posr_out)
{
	/* hist should be initialized array of length = 2*nhist+1 */
	gr_type min = data[0].ratio;
	gr_type max = data[ndata - 1].ratio;
	int i, pos, len = 2 * nhist + 1, val, posl, posr;
	gr_type vall, valr;
	gr_type delta = (max - min) / len;
	gr_type mult = 1. / delta;
	gr_type diff;

	for(i = 0; i < ndata; i++)
	{
		int val = ((data[i].flag == 0) ? 1 : -1);
		diff = data[i].ratio - min;
		pos=(int)(diff * mult);
		if (pos == len)
		{
			pos--;
		}
		hist[pos] += val;
	}
	val = -1;
	pos=0;
	for (i = 0; i < len; i++)
	{
		if (abs(hist[i]) > val) 
		{
			pos = i;
			val = abs(hist[i]);
		}
	}
	vall = min + (pos+0.5) * delta;
	valr = vall + (max - min) / 4;
	vall = vall - (max - min) / 4;
	posl = 0;
	posr = 0;
	for(i = 0; i < ndata; i++)
	{
		if (data[i].ratio < vall)
		{
			posl++;
		}
		if (data[i].ratio < valr)
		{
			posr++;
		}
	}
	posl++;
	if (posr==ndata)
	{
		posr--;
	}
	
	*posl_out = posl;
	*posr_out = posr;  
	
}



ipix_ptr *my_bsearch1(unsigned int value, ipix_ptr *iptr,
							               unsigned int n)
/* The binary search routine  
 Returns the pointer to the idx element of the array such as 
 arr[idx-1] < value <= arr[idx] and 
 or 0 if arr[last_element] < value
*/
{
	ipix_ptr *iptr_cur = iptr;
	unsigned int cur_n = n, t = n;
 

	/* Here we do the binary search on iptr array 
	 * On each step iptr_cur is the pointer to the part of the array 
	 * we are working on.
	 * t -- is the length of the part of the array we working on
	 * cur_n is always t/2 (!integer division!)
	 */   
	while(1)
	{
		cur_n = t / 2;


		/* We stop the loop when the array is 1 or 2 elements long */
		if (t <= 2) 
		{
			break;
		}

		if (iptr_cur[cur_n].ipix < value)
		{
			iptr_cur += (cur_n + 1);
			t -= (cur_n + 1);
		}
		else
		{
			t = cur_n + 1;
		}
	}
	
	if (t == 1)
	/* When at the end we have one-element array */
	{
		if (iptr_cur[0].ipix < value)
		{
			if ((iptr_cur - iptr) == (n - 1)) 
			{
				return 0;
			}
			else 
			{
			/* Should never be reached */
				return iptr_cur + 1;
			} 
		}
		else
		{
			return iptr_cur;
		}
	}
	else 
	/* When at the end we have two-element array */
	{
		if (iptr_cur[0].ipix < value)
		{
			if (iptr_cur[1].ipix < value)
			{
				if ((iptr_cur - iptr) == (n - 2))
				{
					return 0;
				}
				else
				{
				/* Should never be reached */
					return iptr_cur + 2;
				}
			}
			else
			{
				return iptr_cur + 1;
			}
		}
		else
		{
			return iptr_cur;
		}
	}
}

ipix_ptr *my_bsearch2(unsigned int value, ipix_ptr *iptr,
							               unsigned int n)
/* The binary search routine
 Returns the pointer to the idx element of the array such as 
 arr[idx] <= value < arr[idx+1]
 or 0 if value < arr[0]
*/
{
	ipix_ptr *iptr_cur = iptr;
	unsigned int cur_n = n, t = n;

	/* Here we do the binary search on iptr array 
	 * On each step iptr_cur is the pointer to the part of the array 
	 * we are working on.
	 * t -- is the length of the part of the array we working on
	 * cur_n is always t/2 (!integer division!)
	 */   

	while(1)
	{

		cur_n = t/ 2;

		if (t <= 2) 
		{
			break;
		}

		if (iptr_cur[cur_n].ipix <= value)
		{
			iptr_cur += cur_n;
			t -= cur_n;
		}
		else
		{
			t = cur_n;
		}
	}

	if (t == 1)
	/* When at the end we have one-element array */
	{
		if (iptr_cur[0].ipix > value)
		{
			if (iptr_cur == iptr) 
			{
				return 0;
			}
			else 
			{
				/* Should never be reached */
				return iptr_cur - 1;
			}
		}
		else
		{
			return iptr_cur;
		}
	}
	else
	/* When at the end we have two-element array */
	{
		if (iptr_cur[1].ipix <= value)
		{
			return iptr_cur+1;
		}
		else
		{
			if (iptr_cur[0].ipix <= value)
			{
				return iptr_cur;
			}
			else
			{
				if (iptr_cur == iptr) 
				{
					return 0;
				}
				else 
				{
					return iptr_cur - 1;
				}
			}
		}
	}
}



void process_std_list (struct gr_xy *xy_data, int nstars,
						struct std_list **out_data, void **ipix_ptr_arr,
						int *ntriangles)
{
	gr_type d12, d23, d31, dx12, dy12, dx23, dy23, dx31, dy31, r, r1, c, perim;
	struct gr_xy xy1, xy2, xy3;
	unsigned int ipix, nside = 65536, r_int, c_int;
	extern unsigned int ____xbits[], ____ybits[];
	struct std_list *cur_out_data;
	ipix_ptr *cur_ipix_ptr;
	char orient;
	short int id, id2 ,id3;
	int i, j, k;
	*ntriangles = 0 ;

	*ipix_ptr_arr = malloc((nstars * (nstars - 1) * (nstars - 2)) / 6 *
		sizeof(ipix_ptr));
	cur_ipix_ptr = (ipix_ptr *)(*ipix_ptr_arr);
	*out_data = malloc((nstars * (nstars - 1) * (nstars - 2)) / 6 *
		sizeof(struct std_list));
	cur_out_data = (struct std_list *)(*out_data);
	

	/* The Processing loop, computing the parameters of the triangles */
	for(i = 0; i < nstars; i++)
	{
		xy1 = xy_data[i];
		for(j = i + 1; j < nstars; j++)
		{
			xy2 = xy_data[j];
			dx12 = xy1.x - xy2.x;
			dy12 = xy1.y - xy2.y;
			
			d12 = dx12 * dx12 + dy12 * dy12;      
			for(k = j + 1; k < nstars; k++)
			{
				xy3 = xy_data[k];

				dx23 = xy2.x - xy3.x;
				dy23 = xy2.y - xy3.y;

				d23 = dx23 * dx23 + dy23 * dy23;      
				dx31 = xy3.x - xy1.x;
				dy31 = xy3.y - xy1.y;
				
				d31 = dx31 * dx31 + dy31 * dy31;
				

				/* In this loop 
				 * r -- is the ratio between the largest and smallest sides of the tr.
				 * c -- is the cosine of angle between those sides
				 * id -- is the id of the star lying in the vertice between smallest
				 *       and largest sides 
				 * id2 -- is the id of the second star lying on smallest side 
				 * id3 -- is the id of the third star lying on the largest side
				 * orient -- is 0 or 1, depending on the orientation of the triangle
				 */
				if (d12 <= d23)
				{
					if (d23 <= d31)
					{
						/* The case of d12<=d23<=d31 */
						r = d31 / d12;
						c = (d31 + (d12 - d23)) / (2 * sqrt(d31 * d12));
						//xy0 = xy1;
						id = i;
						id2 = j;
						id3 = k;
						orient = TR_OR(dx31, dy31, dx12, dy12);
					}
					else
					{
						if (d12 <= d31)
						{
							/* The case of d12<=d31<d23 */
							r = d23 / d12;
							c = (d23 + (d12 - d31)) / (2 * sqrt(d23 * d12));          
							//xy0 = xy2;
							id = j;
							id2 = i;
							id3 = k;
							orient = TR_OR(dx23, dy23, dx12, dy12);
						}
						else
						{
							/* The case of d31<d12<=d23 */
							r = d23 / d31;            
							c = (d23 + (d31 - d12))/(2 * sqrt(d23 * d31));
							//xy0 = xy3;
							id = k;
							id2 = i;
							id3 = j;
							orient = TR_OR(dx23, dy23, dx31, dy31);
						}                    
					}
				}
				else
				{
					if (d12 <= d31)
					{
						/* The case of d23<d12<=d31 */
						r = d31 / d23;
						c = (d31 + (d23 - d12))/(2 * sqrt(d31 * d23));
						//xy0 = xy3;
						id = k;
						id2 = j;
						id3 = i;
						orient = TR_OR(dx31, dy31, dx23, dy23);
					}
					else
					{
						if (d23 <= d31)
						{
							/* The case of d23<=d31<d12 */
							r = d12 / d23;
							c = (d12 + (d23 - d31))/(2 * sqrt(d12 * d23));
							//xy0 = xy2;
							id = j;
							id2 = k;
							id3 = i;
							orient = TR_OR(dx12, dy12, dx23, dy23);
						}
						else
						{
							/* The case of d31<d23<d12 */
							r = d12 / d31;            
							c = (d12 + (d31 - d23))/(2 * sqrt(d12 * d31));          
							//xy0 = xy1;
							id = i;
							id2 = k;
							id3 = j;
							orient = TR_OR(dx12, dy12, dx31, dy31);
						}                    
					}        
				}
				
				r = sqrt(r);
				perim = d12 + d23 + d31;
				
				if (r >= R_MAX_THRESH)
				{
					continue;
				}
				
				(*ntriangles) += 1;
				
				r1 = (r - 1) / (R_MAX_THRESH - 1 );
				
				r_int = (int) (r1 * nside);
				c_int = (int) (c * nside);
				ipix = ____xbits[r_int] + ____ybits[c_int];
				
			 
				cur_ipix_ptr->ipix = ipix;
				cur_ipix_ptr->ptr = cur_out_data;
				cur_out_data->c = c;
				cur_out_data->r = r;
				cur_out_data->perim = perim;
				cur_out_data->orient = orient;
				//cur_out_data->xy = xy0;
				cur_out_data->id = id;
				cur_out_data->id2 = id2;
				cur_out_data->id3 = id3;
				cur_out_data++;
				cur_ipix_ptr++;
			}
		}  
	}
	
	
	qsort(*ipix_ptr_arr, *ntriangles, sizeof(ipix_ptr), &ipix_compare);

}


void process_obs_list (struct gr_xy *xy_data, int nstars,
						struct obs_list **out_data, int *ntriangles,
						gr_type eps)
{
	gr_type d12, d23, d31, dx12, dy12, dx23, dy23, dx31, dy31, r, c, perim, t, e_c, e_r;
	gr_type eps2;
	short int id, id2, id3;
	struct gr_xy xy1, xy2, xy3;
	struct obs_list *cur_out_data ;
	char orient;

	int i, j, k;
	*ntriangles = 0;

	eps2 = eps * eps;

	*out_data = malloc((nstars* (nstars - 1) * (nstars - 2)) / 6 *
				sizeof(struct obs_list));
	cur_out_data = (struct obs_list *)(*out_data);
	
	for(i = 0; i < nstars; i++)
	{
		xy1 = xy_data[i];
		for(j = i + 1; j < nstars; j++)
		{
			xy2 = xy_data[j];
			dx12 = xy1.x - xy2.x;
			dy12 = xy1.y - xy2.y;
			
			d12 = dx12 * dx12 + dy12 * dy12;
			for(k = j + 1; k < nstars; k++)
			{
				xy3=xy_data[k];

				dx23 = xy2.x - xy3.x;
				dy23 = xy2.y - xy3.y;

				d23 = dx23 * dx23 + dy23 * dy23;      
				dx31 = xy3.x - xy1.x;
				dy31 = xy3.y - xy1.y;
				
				d31 = dx31 * dx31 + dy31 * dy31;
				
				/* In this loop 
				 * r -- is the ratio between the largest and smallest sides of the tr.
				 * c -- is the cosine of angle between those sides
				 * id -- is the id of the star lying in the vertice between smallest
				 *       and largest sides 
				 * id2 -- is the id of the second star lying on smallest side 
				 * id3 -- is the id of the third star lying on the largest side
				 * orient -- is 0 or 1, depending on the orientation of the triangle
				 */
				if (d12 <= d23)
				{
					if (d23 <= d31)
					{
						/* The case of d12<=d23<=d31 */
						r = d31 / d12;
						t = sqrt(d31 * d12);
						c = (d31 + (d12 - d23)) / (2 * t);
						t = (1 / d31 + 1 / d12 - c / t);
						//xy0 = xy1;
						id = i;
						id2 = j;
						id3 = k;
						orient = TR_OR(dx31, dy31, dx12, dy12);
					}
					else
					{
						if (d12 <= d31)
						{
							/* The case of d12<=d31<d23 */
							r = d23 / d12;
							t = sqrt(d23 * d12);
							c = (d23 + (d12 - d31)) / (2 * t);
							t = (1 / d23 + 1 / d12 - c / t);
							//xy0 = xy2;
							id = j;
							id2 = i;
							id3 = k;
							orient = TR_OR(dx23, dy23, dx12, dy12);
						}
						else
						{
							/* The case of d31<d12<=d23 */
							r = d23 / d31;            
							t = sqrt(d23 * d31);
							c = (d23 + (d31 - d12)) / (2 * t);          
							t = (1 / d23 + 1 / d31 - c / t);
							//xy0 = xy3;
							id = k;
							id2 = i;
							id3 = j;
							orient = TR_OR(dx23, dy23, dx31, dy31);
						}                    
					}
				}
				else
				{
					if (d12 <= d31)
					{
						/* The case of d23<d12<=d31 */
						r = d31 / d23;
						t = sqrt(d31 * d23);
						c = (d31 + (d23 - d12)) / (2 * t);
						t = (1 / d31 + 1 / d23 - c / t);
						//xy0 = xy3;
						id = k;
						id2 = j;
						id3 = i;
						orient = TR_OR(dx31, dy31, dx23, dy23);
					}
					else
					{
						if (d23 <= d31)
						{
							/* The case of d23<=d31<d12 */
							r = d12 / d23;
							t = sqrt(d12 * d23);
							c = (d12 + (d23 - d31)) / ( 2 * t);
							t = (1 / d12 + 1 / d23 - c / t);
							//xy0 = xy2;
							id = j;
							id2 = k;
							id3 = i;
							orient = TR_OR(dx12, dy12, dx23, dy23);
						}
						else
						{
							/* The case of d31<d23<d12 */
							r = d12 / d31;            
							t = sqrt(d12 * d31);
							c = (d12 + (d31 - d23)) / (2 * t);
							t = (1 / d12 + 1 / d31 - c / t);
							//xy0 = xy1;
							id = i;
							id2 = k;
							id3 = j;
							orient = TR_OR(dx12, dy12, dx31, dy31);
						}                    
					}        
				}

				r = sqrt(r);

				if (r >= R_MAX_THRESH)
				{
					continue;
				}
				

				e_r = sqrt (2 * r * r * eps2 * t);
				e_c = sqrt (2 * (1 - c * c) * eps2 * t + 3 * c * c *
						eps2 * eps2 * t * t);
							  
				perim = d12 + d23 + d31;

				(*ntriangles) += 1;
				
				//r1 = (r - 1) / (R_MAX_THRESH - 1);
				//r_int = (int) (r1 * nside);
				//c_int = (int) (c * nside);
				//ipix = ____xbits[r_int] + ____ybits[c_int];
				
				cur_out_data->c = c;
				cur_out_data->r = r;
				cur_out_data->e_c = e_c;
				cur_out_data->e_r = e_r;
				cur_out_data->perim = perim;
				cur_out_data->orient = orient;
				//cur_out_data->xy = xy0;
				cur_out_data->id = id;
				cur_out_data->id2 = id2;
				cur_out_data->id3 = id3;
				cur_out_data++;
			}
		}  
	}
}


int match_lists(struct std_list *slist, struct obs_list *olist,
				ipix_ptr *iptr, int nstdtriangles, int nobstriangles)
{
	unsigned int i, n0, nside = 65536, ipixstack1[4], ipixstack2[4], xi, yi, n1,
			ixmax, ixmin, iymax, iymin, nistack, j;

	extern unsigned int ____xbits[], ____ybits[];
	
	ipix_ptr *ptr1, *ptr2, *ptr;
	gr_type c_obs, r_obs, e_c_obs, e_r_obs, r1_obs, e_r1_obs, r_std, r1_std,
					c_std, norm, norm_r, norm_c, size;
	int count_matched = 0;

	
	for(i = 0; i < nobstriangles; i++)
	{
		r_obs = olist[i].r;
		c_obs = olist[i].c;
		e_r_obs = olist[i].e_r;
		e_c_obs = olist[i].e_c;

		r1_obs = (r_obs - 1) / (R_MAX_THRESH - 1);
		e_r1_obs = e_r_obs / (R_MAX_THRESH - 1);
		
		size = 2 * ((e_r1_obs > e_c_obs) ? e_r1_obs : e_c_obs);
		
		if (size * nside < 1)
		{
			size = ((gr_type)1) / nside;
		}
		
		n0 = 1 << ((int)(-ceil(log(size)/log(2))));

		n1 = nside / n0;
		
		ixmin = (r1_obs - e_r1_obs) * n0;
		ixmax = (r1_obs + e_r1_obs) * n0;
		iymin = (c_obs - e_c_obs) * n0;
		iymax = (c_obs + e_c_obs) * n0;

		if (iymin == iymax)
		{
			if (ixmin == ixmax)
			{
				xi = (int)(ixmin * n1);
				yi = (int)(iymin * n1);
				ipixstack1[0] = ____xbits[xi] + ____ybits[yi];
				ipixstack2[0] = ipixstack1[0] + n1 * n1 - 1;
				nistack = 1;   
			}
			else
			{   
				xi = (int)(ixmin * n1);
				yi = (int)(iymin * n1);
				ipixstack1[0] = ____xbits[xi] + ____ybits[yi];
				ipixstack2[0] = ipixstack1[0] + n1 * n1 - 1;
				xi = (int)(ixmax * n1);
				yi = (int)(iymin * n1);
				ipixstack1[1] = ____xbits[xi] + ____ybits[yi];
				ipixstack2[1] = ipixstack1[1] + n1 * n1 - 1;
				nistack = 2;   
			}
		}  
		else
		{
			if (ixmin == ixmax)
			{
				xi = (int)(ixmin * n1);
				yi = (int)(iymin * n1);
				ipixstack1[0] = ____xbits[xi] + ____ybits[yi];
				ipixstack2[0] = ipixstack1[0] + n1 * n1 - 1;
				xi = (int)(ixmin * n1);
				yi = (int)(iymax * n1);
				ipixstack1[1] = ____xbits[xi] + ____ybits[yi];
				ipixstack2[1] = ipixstack1[1] + n1 * n1 - 1;
				nistack = 2;
			}   
			else
			{
				xi = (int)(ixmin * n1);
				yi = (int)(iymin * n1);
				ipixstack1[0] = ____xbits[xi] + ____ybits[yi];
				ipixstack2[0] = ipixstack1[0] + n1 * n1 - 1;
				xi = (int)(ixmin * n1);
				yi = (int)(iymax * n1);
				ipixstack1[1] = ____xbits[xi] + ____ybits[yi];
				ipixstack2[1] = ipixstack1[1] + n1 * n1 - 1;
				xi = (int)(ixmax * n1);
				yi = (int)(iymin * n1);
				ipixstack1[2] = ____xbits[xi] + ____ybits[yi];
				ipixstack2[2] = ipixstack1[2] + n1 * n1 - 1;
				xi = (int)(ixmax * n1);
				yi = (int)(iymax * n1);
				ipixstack1[3] = ____xbits[xi] + ____ybits[yi];
				ipixstack2[3] = ipixstack1[3] + n1 * n1 - 1;
				nistack = 4;
			}
		}
		
		/* Now we setup the query    xx <=   ipix  <= xx + n1 * n1-1 */
		
		norm = 1000;
		olist[i].matched_ptr = 0;
		
		for (j = 0; j < nistack; j++)
		{
			ptr1 = my_bsearch1(ipixstack1[j], iptr, nstdtriangles);

			if (ptr1 == 0) 
			{
				continue;
			}

			ptr2 = my_bsearch2(ipixstack2[j], ptr1, nstdtriangles - (ptr1 - iptr));     

			if (ptr2 == 0) 
			{
				continue;
			}
						
			for(ptr = ptr1; ptr <= ptr2; ptr++)
			{
				r_std = ptr->ptr->r;
				c_std = ptr->ptr->c;
				r1_std = (r_std - 1) / (R_MAX_THRESH - 1);
				
				norm_r = fabs((r1_std - r1_obs) / e_r1_obs);
				norm_c = fabs((c_std - c_obs) / e_c_obs);
				
				if ((norm_c < 1) && (norm_r < 1))
				{
					if ((norm_c + norm_r) < norm)
					{
						norm = norm_c + norm_r;
						olist[i].matched_ptr = ptr->ptr;
						count_matched++;
					}
				}
			}
		}  
	}
	
	if (count_matched == 0)
	{
		fprintf(stderr,"ERROR: No triangles matched!!!\n");
		return -1;
	}
	return 0;
}


int get_transform(struct obs_list *olist, struct gr_xy *obs_xy,
							     struct gr_xy *std_xy, int nobstriangles, int nobsstars,
							     gr_matrix *m, gr_vector *v)
{
	struct gr_xy *sxy, *oxy;
	gr_type *operim, *sperim, *ratio, *ratio1, mean_ratio, mean_ratio1, st_dev,
					t, thresh, max, t1, tx, ty, tx1, ty1, txx, txy, txx1, txy1,
					tyy, tyx1, tyy1, A, B, C, D, vall, valr;
	int nmatch, i, m0, m1, j, cur_n, n_final_triangles, mf, mt;
	char *oorient, *sorient, orient;
	double_and_char *ratio_and_flag;
	char *flag;
	struct three_ints *star_list;
	struct obs_list **optr;
	int *hist, nhist = 10, posl, posr, posl1, posr1, ret_val;
	
	sxy = malloc(sizeof(struct gr_xy) * nobstriangles);
	oxy = malloc(sizeof(struct gr_xy) * nobstriangles);
	optr = malloc(sizeof(struct obs_list *) * nobstriangles);
	
	/* The list of perimeters of observed triangles */
	operim = malloc(sizeof(gr_type) * nobstriangles);

	/* The list of perimeters of standart triangles */
	sperim = malloc(sizeof(gr_type) * nobstriangles);

	/* The list of ratios of triangles */
	ratio = malloc(sizeof(gr_type) * nobstriangles);

	/* The list of orientations of standart triangles */
	sorient = malloc(nobstriangles);

	/* The list of orientations of observed triangles */
	oorient = malloc(nobstriangles);

	/* The space for the histogram used in mode determination*/
	hist = malloc((2 * nhist + 1) * sizeof(int));



	for(i = 0, mean_ratio = 0, nmatch = 0; i < nobstriangles; i++)
	{
		if (!(olist[i].matched_ptr)) continue;
		
		operim[nmatch] = olist[i].perim;
		oxy[nmatch] = obs_xy[olist[i].id]; 
		sperim[nmatch] = olist[i].matched_ptr->perim;
		sxy[nmatch] = std_xy[olist[i].matched_ptr->id];
		optr[nmatch] = olist + i;
		oorient[nmatch] = olist[i].orient;
		sorient[nmatch] = olist[i].matched_ptr->orient;
		ratio[nmatch] = 0.5 * log10(operim[nmatch] / sperim[nmatch]);
		mean_ratio += ratio[nmatch]; 
		
#ifdef GMATCH_DEBUG
		fprintf(stderr, "%d %f\n", nmatch, ratio[nmatch]);
#endif
		
		nmatch++;
	}
	
	mean_ratio /= nmatch;

	free(operim);
	free(sperim); 

	oxy = realloc(oxy, sizeof(struct gr_xy) * nmatch);
	sxy = realloc(sxy, sizeof(struct gr_xy) * nmatch);
	optr = realloc(optr, sizeof(struct obs_list *) * nmatch);
	
	oorient = realloc(oorient, nmatch);
	sorient = realloc(sorient, nmatch);
	ratio = realloc(ratio, sizeof(gr_type) * nmatch);
		
	/* this array will contain the sorted array of ratios of 
		 perimeters of triangles  */
	ratio1 = malloc(sizeof(gr_type) * nmatch);
	ratio_and_flag = malloc(sizeof(double_and_char) * nmatch);

	/* The flag will be used in false match removal 
		 (to flag the bad triangle matches)*/
	flag = calloc(nmatch, 1);

	
	/* That list will be used in final matching process */
	star_list = calloc(3 * nmatch, sizeof(struct three_ints));


	
	for(i = 0; i < nmatch; i++)
	{
		ratio1[i] = ratio[i];
		ratio_and_flag[i].ratio = ratio[i];
		ratio_and_flag[i].flag = (sorient[i]==oorient[i]);
	}
	qsort(ratio1, nmatch, sizeof(gr_type), gr_type_compare);
	qsort(ratio_and_flag, nmatch, sizeof(double_and_char), double_and_char_compare);

#ifdef GMATCH_DEBUG
	for(i = 0; i < nmatch; i++)
	{
		fprintf(stderr,"Y %d %f\n",sorient[i]==oorient[i],ratio[i]);
	}
#endif
	
	posl = 0; 
	posr = nmatch - 1;
	for(i = 0; (i < 8) && ((posr - posl) > 10) && ((ratio_and_flag[posr].ratio-ratio_and_flag[posl].ratio)>0.03); i++)
	{
		memset(hist, 0, (2 * nhist + 1) * sizeof(int));
		//mode_hist(ratio1 + posl, posr - posl + 1, hist, nhist, &posl1, &posr1);
		mode_hist1(ratio_and_flag + posl, posr - posl + 1, hist, nhist, &posl1, &posr1);
		posr = posr1 + posl;
		posl = posl1 + posl;
		
#ifdef GMATCH_DEBUG
		fprintf(stderr, "T %d %f %f %d %d\n",i, ratio1[posl],ratio1[posr], posl, posr);
#endif
	}
	vall = ratio1[posl];
	valr = ratio1[posr];

	for(i = posl, mean_ratio = 0; i <= posr; i++)
	{
		mean_ratio+=ratio1[i];
	}
	mean_ratio /= (posr - posl + 1);

	for(i = 0, st_dev = 0, m0 = 0, m1 = 0; i < nmatch; i++)
	{
		if ((ratio[i] < vall) || (ratio[i] > valr)) 
		{
			continue;
		}
#ifdef GMATCH_DEBUG
		fprintf(stderr, "Z %f\n",ratio[i]);
#endif

		t = (ratio[i] - mean_ratio);
		st_dev += t * t;
		if (sorient[i] != oorient[i]) 
		{
			m0++;
		}
		else
		{
			m1++;
		}
	}
	
	st_dev = sqrt(st_dev / (posr-posl+1));
	
#ifdef GMATCH_DEBUG    
		fprintf(stderr,"U %d %d %d %f %f\n", nmatch, m0, m1, mean_ratio, st_dev);
#endif
		
	
	/* We perform 10 iterations to reject the false matches */
	for(j = 0; j < 10; j++)
	{


		mt = ((m1 - m0) > 0) ? (m1 - m0) : (m0 - m1);
		mf = m1 + m0 - mt;
		if (mf > mt)
		{
			thresh = 1;
		}
		else
		{
			if (0.1 * mt > mf)
			{
				thresh = 3;
			}
			else
			{
				thresh = 2;
			}
		}

#ifdef GMATCH_DEBUG    
		fprintf(stderr,"U %d %d %d %d %d %f %f %f\n", nmatch, m0, m1, mf, mt, mean_ratio, st_dev, thresh);
#endif
		
		for(i = 0, mean_ratio1 = 0, cur_n = 0; i < nmatch; i++)
		{
			if (flag[i])
			{
				continue;
			}
			
			if (fabs(ratio[i] - mean_ratio) > thresh * st_dev)
			{
				flag[i] = 1;
				continue;
			}
			
			mean_ratio1 += ratio[i];
			cur_n++;      
		}
		
		mean_ratio = mean_ratio1 / cur_n;
		
		for(i = 0, m0 = 0, m1 = 0, st_dev = 0; i < nmatch; i++)
		{
			if (flag[i])
			{
				continue;
			}

			t = (ratio[i] - mean_ratio);
			st_dev += t * t;
			if (sorient[i] != oorient[i])
			{
				m0++;
			}
			else
			{
				m1++;
			}
		}
		st_dev = sqrt(st_dev / cur_n); 
	}
	
	if (m1 > m0)
	{
		for(i = 0; i < nmatch; i++)
		{
			if (sorient[i] != oorient[i])
			{
				flag[i] = 1;
			}
		}
		orient = 1;
	}
	else
	{
		for(i = 0; i < nmatch; i++)
		{
			if (sorient[i] == oorient[i])
			{
				flag[i] = 1;
			}
		}  
		orient = 0;
	}
	
	
	
	for(i = 0, n_final_triangles = 0; i < nmatch; i++)
	{
		if (flag[i])
		{
			continue;
		}
		/*
		fprintf(stderr,"%d %d %d %f %f %f %f\n",i,optr[i]->id, optr[i]->matched_ptr->id,
						obs_xy[optr[i]->id].x, 
						obs_xy[optr[i]->id].y, std_xy[optr[i]->matched_ptr->id].x, 
						std_xy[optr[i]->matched_ptr->id].y);
		fprintf(stderr,"%d %d %d %f %f %f %f\n",i,optr[i]->id3, optr[i]->matched_ptr->id3,
						obs_xy[optr[i]->id3].x, 
						obs_xy[optr[i]->id3].y, std_xy[optr[i]->matched_ptr->id3].x, 
						std_xy[optr[i]->matched_ptr->id3].y);
		fprintf(stderr,"%d %d %d %f %f %f %f\n",i,optr[i]->id2, optr[i]->matched_ptr->id2,
						obs_xy[optr[i]->id2].x, 
						obs_xy[optr[i]->id2].y, std_xy[optr[i]->matched_ptr->id2].x, 
						std_xy[optr[i]->matched_ptr->id2].y);
		*/

		star_list[3*n_final_triangles].l2 = optr[i]->id;
		star_list[3*n_final_triangles].l3 = optr[i]->matched_ptr->id;
		star_list[3*n_final_triangles+1].l2 = optr[i]->id2;
		star_list[3*n_final_triangles+1].l3 = optr[i]->matched_ptr->id2;
		star_list[3*n_final_triangles+2].l2 = optr[i]->id3;
		star_list[3*n_final_triangles+2].l3 = optr[i]->matched_ptr->id3;
		n_final_triangles++;
	}

	/*
	for(i = 0; i < 3 * n_final_triangles; i++)
	{
		fprintf(stderr,"X %d %d %d %d\n",i,star_list[i].l1,star_list[i].l2,star_list[i].l3);
	}
	*/

	qsort(star_list, 3 * n_final_triangles, sizeof(struct three_ints), &three_ints_compare1);

	/*
	for(i = 0; i < 3 * n_final_triangles; i++)
	{
		fprintf(stderr,"A %d %d %d\n",star_list[i].l1,star_list[i].l2,star_list[i].l3);
	}
	*/

	for(i = 0, m0 = 0; i < 3 * n_final_triangles; i++)
	{
		if (m0 == 0)
		{
			m0 = 1;
			m1 = i;
			star_list[i].l1+=1;
		}
		else 
		{
			if (three_ints_compare1(star_list + m1, star_list + i) == 0)
			{
				star_list[m1].l1 += 1;
				continue;
			
			}
			else 
			{
				star_list[i].l1 += 1;
				m1 = i;
				m0 = 1;
			}
		}
	}

	/*
	for(i = 0;i<3 * n_final_triangles; i++)
	{
		fprintf(stderr,"B %d %d %d\n",star_list[i].l1,star_list[i].l2,star_list[i].l3);
	}
	*/
/*
	if (n_final_triangles == nmatch) 
	{
		fprintf(stderr,"ERROR: Triangles do not match!!!\n");
		return -1;
	}
	*/
	qsort(star_list, 3 * n_final_triangles, sizeof(struct three_ints), &three_ints_compare);
	

#ifdef GMATCH_DEBUG  
	fprintf(stderr,"\n");
	for(i = 0; i < 3 * n_final_triangles; i++)
	{
		fprintf(stderr,"C %d %d %d\n", star_list[i].l1, star_list[i].l2, star_list[i].l3);
	}
#endif  

	max = star_list[0].l1;

	if (max < 1) 
	{
		fprintf(stderr,"ERROR: Triangles do not match!!!\n");
		ret_val = -1;
		goto GET_TRANSFORM_END;
	}

	tx = 0;
	ty = 0;
	tx1 = 0;
	ty1 = 0;
	txx = 0;
	txy=0;
	txx1 = 0;
	txy1 = 0;
	tyy = 0;
	tyx1 = 0;
	tyy1 = 0;
	t1 = 0;

	for (i = 0; i < 3 * n_final_triangles && (star_list[i].l1 >= (max / 2)); i++)
	{
		
#ifdef GMATCH_DEBUG
		fprintf(stderr,"%d %d %d %f %f %f %f\n",star_list[i].l1, star_list[i].l2,
						star_list[i].l3, obs_xy[star_list[i].l2].x, 
						obs_xy[star_list[i].l2].y, std_xy[star_list[i].l3].x, 
						std_xy[star_list[i].l3].y);
#endif
		
		t1   += 1;  
		tx   += obs_xy[star_list[i].l2].x;
		ty   += obs_xy[star_list[i].l2].y;
		tx1  += std_xy[star_list[i].l3].x;
		ty1  += std_xy[star_list[i].l3].y;
		txx  += obs_xy[star_list[i].l2].x * obs_xy[star_list[i].l2].x;
		txy  += obs_xy[star_list[i].l2].x * obs_xy[star_list[i].l2].y;
		txx1 += obs_xy[star_list[i].l2].x * std_xy[star_list[i].l3].x;
		txy1 += obs_xy[star_list[i].l2].x * std_xy[star_list[i].l3].y;
		tyy  += obs_xy[star_list[i].l2].y * obs_xy[star_list[i].l2].y;
		tyx1 += obs_xy[star_list[i].l2].y * std_xy[star_list[i].l3].x;
		tyy1 += obs_xy[star_list[i].l2].y * std_xy[star_list[i].l3].y;

	}

#ifdef GMATCH_DEBUG
	fprintf(stderr,"OOOOO %d\n",orient);
#endif  
//  orient = 1;
	if (orient == 0) 
	{
		get_regression2_1(t1, tx, ty, tx1, ty1, txx, txx1, txy1, tyy, tyx1,
							        tyy1, &A, &B, &C, &D);
	}
	else
	{
		get_regression2  (t1, tx, ty, tx1, ty1, txx, txx1, txy1, tyy, tyx1,
							        tyy1, &A, &B, &C, &D);
	}

	m->t[0] = A;
	m->t[1] = B;
	if (orient == 0)
	{
		m->b[0] = B;
		m->b[1] = -A;

	}
	else
	{
		m->b[0] = -B;
		m->b[1] = A;
	}
	v->t[0] = C;
	v->b[0] = D;

	st_dev = 0;
	t1 = 0;
	for (i = 0; i < 3 * n_final_triangles && (star_list[i].l1 >= (max / 2)); i++)
	{
		
		t1 += 1;  
/*
		tx = obs_xy[star_list[i].l2].x - m->t[0]*std_xy[star_list[i].l3].x- m->t[1]*std_xy[star_list[i].l3].y - v->t[0]; 
		ty = obs_xy[star_list[i].l2].y - m->b[0]*std_xy[star_list[i].l3].x - m->b[1]*std_xy[star_list[i].l3].y - v->b[0];
		
		tx = obs_xy[star_list[i].l2].x - m->t[0] * (std_xy[star_list[i].l3].x - v->t[0]) -
			m->t[1] * (std_xy[star_list[i].l3].y - v->b[0]);
		ty = obs_xy[star_list[i].l2].y - m->b[0] * (std_xy[star_list[i].l3].x - v->t[0]) -
			m->b[1] * (std_xy[star_list[i].l3].y - v->b[0]);
*/
		tx = std_xy[star_list[i].l3].x - m->t[0] * obs_xy[star_list[i].l2].x -
			m->t[1] *  obs_xy[star_list[i].l2].y - v->t[0];
		ty = std_xy[star_list[i].l3].y - m->b[0] * obs_xy[star_list[i].l2].x -
			m->b[1] *  obs_xy[star_list[i].l2].y - v->b[0];

		st_dev += (tx*tx+ty*ty);
	}
	st_dev = sqrt(st_dev/t1);
	fprintf(stderr,"RMS: %f",st_dev);


/*               !!  TODO  !!

 Here I should insert the code to compute the RMS of transformation 
 
 */
 

#ifdef GMATCH_DEBUG
	fprintf(stderr,"%f %f %f %f %f %f\n",m->t[0], m->t[1], m->b[0], m->b[1],
					v->t[0], v->b[0]);
#endif

	get_regression(t1, tx, ty, tx1, txx, txx1, txy, tyx1, tyy, &A, &B, &C);
	get_regression(t1, tx, ty, ty1, txx, txy1, txy, tyy1, tyy, &A, &B, &C);

#ifdef GMATCH_DEBUG
	fprintf(stderr,"%f %f %f\n", A, B, C);
#endif

	
	ret_val=0;
	
	GET_TRANSFORM_END:
	
	free(oxy);
	free(sxy);
	free(optr);
	free(oorient);
	free(sorient);
	free(ratio);
	free(ratio1);
	free(flag);
	free(hist);
	free(star_list);

	return ret_val;
}


void get_new_wcs(gr_type scale, gr_matrix m, gr_vector c, int xsize, int ysize, gr_matrix *cd, gr_vector *pixref)
{
	gr_matrix one, m_inverted;
	gr_vector v, c0, v1;
	one.t[0] = scale; one.t[1] = 0; one.b[0] = 0; one.b[1] = scale;
	c0.t[0] = (xsize + 1.) / 2.; c0.t[1] = (ysize + 1.) / 2.;
	matrix_multiply(m, one, cd);
	m_inverted = matrix_invert(m);
	v = vector_subtract(c0, c);
	v1 = matrix_vector_multiply(m_inverted, v);
	*pixref = v1; 
}


//double *read_convert_test(char *filename, double ra_cen, double dec_cen, double scale, int xsize, int ysize, int *nstars0);

/* Returns 0 in the case of success */
int run_groth(char *fits_filename, char *obs_list_filename,
							 char *std_list_filename, gr_type ra_cen, gr_type dec_cen,
							 gr_type eps, gr_type *crpix_out, gr_type *crval_out,
							 gr_type *cd_out, int nobsstars, int nstdstars, char use_wcs)
{

	struct gr_xy *obs_xy, *std_xy;

	int i, nobstriangles, nstdtriangles,
			nstars, xsize, ysize;
	struct std_list *std_data;
	struct obs_list *obs_data;
	double *xy;
	void *ipix_ptr;
	FILE *fp;
	gr_matrix m;
	gr_vector v,  pixref;
	gr_matrix cd;
	gr_type scale = 0.0001; /* Can be arbitrary value */
	int status;
	long size[2];
	struct wcsprm our_wcs;
	
	
	if (use_wcs)
	{
		status = fits_wcs_init(fits_filename, &our_wcs, size);
		scale = sqrt(our_wcs.cd[0]*our_wcs.cd[0]+our_wcs.cd[1]*our_wcs.cd[1]);
	}
	else
	{
		status = gmatch_get_image_size(fits_filename, size);  
	}
	
	if (status != 0)
	{
		return -1;
	}  
	
	xsize=size[0];
	ysize=size[1];

	obs_xy = malloc(sizeof(struct gr_xy) * nobsstars);
	
//  double *radec=malloc(2*sizeof(double)*nobsstars);

	nstars = nstdstars;

	if (use_wcs)
	{
		xy = read_and_wcs_convert1(std_list_filename, ra_cen, dec_cen, scale, 
							                 xsize, ysize, &our_wcs, &nstars);
	}
	else
	{
		xy = read_and_wcs_convert(std_list_filename, ra_cen, dec_cen, scale,
							                xsize,  ysize, &nstars);
	}

	nstdstars = nstars;
	
	if (xy == 0)
	{
		free(obs_xy);
		fprintf(stderr,"Strange error!!!\n");
		return -1;
	}

	if (nstdstars <= 2)
	{
		fprintf(stderr,"Not enough stars in the catalogue list!!!\n");
		return -1;
	}
	
	

	std_xy = malloc(sizeof(struct gr_xy) * nstdstars);
	
	for(i = 0; i < nstdstars; i++)
	{
		std_xy[i].x = xy[2 * i];
		std_xy[i].y = xy[2 * i + 1];
		//fprintf(stderr,"%f %f\n", std_xy[i].x,  std_xy[i].y);
	}
	

	free(xy);

	//fp=fopen("obs1.cat","r");


	fp = fopen(obs_list_filename, "r");
	if (!fp) 
	{
		perror(obs_list_filename);
		return -1;
	}

	for (i = 0; !feof(fp) && (i < nobsstars); )
	{
		char buf[101];
		fgets(buf,100,fp);
		if (2 == sscanf(buf, "%lf %lf", &(obs_xy[i].x), &(obs_xy[i].y)))
		{	
			i++;
		}
	}

	nobsstars = i;

	if (nobsstars <= 2)
	{
		fprintf(stderr,"Not enough stars in the observed list!!!\n");
		return -1;
	}


	nobstriangles = (nobsstars * (nobsstars - 1) * (nobsstars - 2)) / 6;

	fclose(fp);

	process_obs_list (obs_xy, nobsstars, &obs_data, &nobstriangles, eps);
	process_std_list (std_xy, nstdstars, &std_data, &ipix_ptr, &nstdtriangles);

	status = match_lists(std_data, obs_data, ipix_ptr, nstdtriangles, nobstriangles);
	free(ipix_ptr);

	if (status != 0)
	{
		return -1;
	}


	/*
	for(i = 0;i < nobstriangles; i++)
	{
		if (obs_data[i].matched_ptr)
		{
			fprintf(stderr,"%d %f %f %f %d %f %f %f\n",obs_data[i].id, obs_data[i].perim,
							obs_xy[obs_data[i].id].x, obs_xy[obs_data[i].id].y, obs_data[i].matched_ptr->id,
							obs_data[i].matched_ptr->perim, std_xy[obs_data[i].matched_ptr->id].x,
							std_xy[obs_data[i].matched_ptr->id].y);  
		}
	}
	*/

	status = get_transform(obs_data, obs_xy, std_xy, nobstriangles, nobsstars, &m, &v);
	
	free(obs_data);
	free(std_data);

	free(obs_xy);
	free(std_xy);

	if (status != 0)
	{
		return -1;
	}
	get_new_wcs(scale, m, v, xsize, ysize, &cd, &pixref);

#ifdef GMATCH_DEBUG
	
	matrix_print(stderr, cd);
	vector_print(stderr, pixref);
	fprintf(stderr, "%.10f %.10f\n",ra_cen, dec_cen);
#endif
	
	crval_out[0] = ra_cen;
	crval_out[1] = dec_cen;
	crpix_out[0] = pixref.t[0];
	crpix_out[1] = pixref.b[0];
	cd_out[0] = cd.t[0];
	cd_out[1] = cd.t[1];
	cd_out[2] = cd.b[0];
	cd_out[3] = cd.b[1];
	return 0;
}


